#include <nlp_filesystem.h>
#include <nlp_config.h>
#include <nlp_macros.h>
#include <gsl/gsl_randist.h>
#include <time.h>
#include <assert.h>
#include <limits>
#include <math.h>
#include "Model.h"
#include "Phrase.h"
#include "Document.h"

unsigned int 	Model::sui_Topics = 0;
double			Model::d_TopicSmoothing = 0;
unsigned int	Model::sui_TopicSamples = 0;
bool			Model::sb_PredictTestDocumentProperties = false;


// -------------------------------------------------------------------
ThreadInfo::ThreadInfo (void)
{
	pp_DocumentsForSampling = NULL;
	gsl_rng_env_setup ();
	gsl_rng_default_seed = (unsigned long) time (NULL);
	p_GslRng = gsl_rng_alloc (gsl_rng_default);

	pd_TypeTopicCountData = new double [Model::sui_Topics * Document::sui_Types];

	pthread_mutex_init (&mtx_Start, NULL);
	pthread_mutex_init (&mtx_Done, NULL);
	pthread_mutex_lock (&mtx_Start);
	pthread_mutex_lock (&mtx_Done);

	b_Done = false;
	pthread_create (&o_Thread, NULL, Model::SampleDocumentLanguageModels, this);
	pthread_detach (o_Thread);
}


ThreadInfo::~ThreadInfo (void)
{
	pthread_mutex_destroy (&mtx_Start);
	pthread_mutex_destroy (&mtx_Done);

	delete[] pd_TypeTopicCountData;
	pd_TypeTopicCountData = (double*)0x1;

	if (NULL != p_GslRng)
		gsl_rng_free (p_GslRng);
	p_GslRng = (gsl_rng*)0x1;

	dq_Documents.clear ();
	if (NULL != pp_DocumentsForSampling)
		delete[] pp_DocumentsForSampling;
	pp_DocumentsForSampling = (Document**)0x1;
}


// -------------------------------------------------------------------
Model::Model (void)
{
	// Initialize GSL random number generator ...	
	gsl_rng_env_setup ();
	gsl_rng_default_seed = (unsigned long) time (NULL);
	p_GslRng = gsl_rng_alloc (gsl_rng_default);

	// initialize model parameters ...				
	sui_Topics = (int)(config)"topic_count";
	sui_TopicSamples = (int)(config)"topic_samples_for_prediction";

	pd_Theta = new double [sui_Topics];
	for (unsigned int i = 0; i < sui_Topics; ++ i)
		pd_Theta [i] = 1 / (double)sui_Topics;

	pp_PhrasesForSampling = NULL;
	sb_PredictTestDocumentProperties = false;
}


// -------------------------------------------------------------------
Model::~Model (void)
{
	if (NULL != pd_Theta)
		delete[] pd_Theta;
	pd_Theta = (double*)0x1;

	if (NULL != p_GslRng)
		gsl_rng_free (p_GslRng);
	p_GslRng = (gsl_rng*)0x1;

	if (NULL != pd_PhiData)
		delete[] pd_PhiData;
	pd_PhiData = (double*)0x1;
	if (NULL != ppd_Phi)
		delete[] ppd_Phi;
	ppd_Phi = (double**)0x1;

	if (NULL != pui_TopicLogData)
		delete[] pui_TopicLogData;
	pui_TopicLogData = (unsigned int*)0x1;

	if (NULL != pp_PhrasesForSampling)
		delete[] pp_PhrasesForSampling;
	pp_PhrasesForSampling = (Phrase**)0x1;
}


// -------------------------------------------------------------------
void Model::Init (void)
{
	// create model state log path
	if (false == Path::Exists ((config)"state_path"))
		if (false == Path::CreatePath ((config)"state_path", 0700))
		{
			cerr << "[ERROR]  Failed to create model state log path : " << (config)"state_path" << endl;
			return;
		}
	// delete old model parameter output files if they exist 
	if (true == Path::Exists ((config)"z_file"))
		Path::RemoveFile ((config)"z_file");
	if (true == Path::Exists ((config)"theta_file"))
		Path::RemoveFile ((config)"theta_file");

	// 														
	InitParameters ();
	InitThreads ();
}


// -------------------------------------------------------------------
void Model::InitThreads (void)
{
	i_ThreadCount = (config)"thread_count";
	cout << "Initializing " << i_ThreadCount << " threads." << endl;

	p_ThreadInfo = new ThreadInfo [i_ThreadCount];

	for (int i = 0; i < i_ThreadCount; ++ i)
	{
		p_ThreadInfo [i].d_AlphaD = d_AlphaD;
		p_ThreadInfo [i].ppd_Phi = ppd_Phi;
	}

	for (unsigned long l = 0; l < Document::sdq_Documents.size (); ++ l)
		p_ThreadInfo [l % i_ThreadCount].dq_Documents.push_back (Document::sdq_Documents [l]);
	for (int i = 0; i < i_ThreadCount; ++ i)
	{
		ThreadInfo* pInfo = &(p_ThreadInfo [i]);
		pInfo->l_DocumentsForSampling = pInfo->dq_Documents.size ();
		pInfo->pp_DocumentsForSampling = new Document* [pInfo->l_DocumentsForSampling];
		for (long j = 0; j < pInfo->l_DocumentsForSampling; ++ j)
			pInfo->pp_DocumentsForSampling [j] = pInfo->dq_Documents [j];
	}
}


// -------------------------------------------------------------------
void Model::InitParameters (void)
{
	// initialize model parameters ...
	d_TopicSmoothing = (config)"topic_smoothing";
	d_Alpha = (config)"cluster_prior";
	d_AlphaD = (config)"topic_prior";
	d_Beta = (config)"lm_prior";

	d_TopicSmoothingScaling = (config)"topic_smoothing_scaling";
	d_AlphaScaling = (config)"cluster_prior_scaling";
	d_AlphaDScaling = (config)"topic_prior_scaling";
	d_BetaScaling = (config)"lm_prior_scaling";

	double dUniformTypeProb = 1 / (double)Document::sui_Types;
	pd_PhiData = new double [Document::sui_Types * Model::sui_Topics];
	for (unsigned long i = 0; i < Document::sui_Types * Model::sui_Topics; ++ i)
		pd_PhiData [i] = dUniformTypeProb;
	ppd_Phi = new double* [Document::sui_Types];
	for (unsigned int i = 0; i < Document::sui_Types; ++ i)
		ppd_Phi [i] = &(pd_PhiData [i * Model::sui_Topics]);

	pui_TopicLogData = new unsigned int [1000 * Phrase::sui_PhraseCount];


	// initialize sampling structure for phrases
	pp_PhrasesForSampling = new Phrase* [Phrase::sui_PhraseCount];
	for (unsigned int i = 0; i < Phrase::sui_PhraseCount; ++ i)
		pp_PhrasesForSampling [i] = Phrase::sdq_Phrases [i];

	
	// initialize phrase documents ...
	ITERATE	(Document_dq_t, Document::sdq_Documents, ite)
	{
		Document* pDoc = *ite;
		for (unsigned int i = 0; i < pDoc->ui_DocumentPhrases; ++ i)
			pDoc->pp_PhraseAnnotation [i]->dq_PhraseDocuments.push_back (pDoc);
	}
	

	// truncate model parameter files ...
	CsvFile csvTheta ((config)"theta_file", ios_base::out);
	CsvFile csvZ ((config)"z_file", ios_base::out);
}


// -------------------------------------------------------------------
void Model::SaveClusterState (void)
{
	// save iteration number 

	CsvFile csvTheta ((config)"theta_state_file", ios_base::out);
	csvTheta.WriteLine (pd_Theta, Model::sui_Topics);
	
	int iPhraseTopics [Phrase::sui_PhraseCount];
	for (unsigned int i = 0; i < Phrase::sui_PhraseCount; ++ i)
		iPhraseTopics [i] = Phrase::sdq_Phrases [i]->ui_Topic;
	
	CsvFile csvTopic ((config)"phrase_topic_state_file", ios_base::out);
	csvTopic.WriteLine (iPhraseTopics, Phrase::sui_PhraseCount);
}


// -------------------------------------------------------------------
void Model::LoadClusterState (void)
{
	// load iteration number ...
	unsigned long ulTopics = Model::sui_Topics;
	CsvFile csvTheta ((config)"theta_state_file");
	csvTheta.ReadLine (pd_Theta, &ulTopics);

	unsigned long ulPhraseCount = Phrase::sui_PhraseCount;
	int iPhraseTopics [Phrase::sui_PhraseCount];
	CsvFile csvTopic ((config)"phrase_topic_state_file");
	if (true == csvTopic.ReadLine (iPhraseTopics, &ulPhraseCount))
	{
		for (unsigned int i = 0; i < Phrase::sui_PhraseCount; ++ i)
			Phrase::sdq_Phrases [i]->ui_Topic = iPhraseTopics [i];
	}
}


// -------------------------------------------------------------------
void Model::SaveLanguageModelState (void)
{
	// save iteration number ...
	
	// save language model ...
	CsvFile csvPhi ((config)"phi_state_file", ios_base::out);
	csvPhi.WriteLine (pd_PhiData, Model::sui_Topics * Document::sui_Types);

	// save document Thetas, R, token topics & token topic sources ...
	CsvFile csvTheta ((config)"thetad_state_file", ios_base::out);
	CsvFile csvR ((config)"r_state_file", ios_base::out);
	CsvFile csvTokenTopic ((config)"token_topic_state_file", ios_base::out);
	CsvFile csvTokenTopicSource ((config)"token_topic_source_state_file", ios_base::out);

	ITERATE (Document_dq_t, Document::sdq_Documents, ite)
	{
		Document* pDocument = *ite;

		csvTheta.WriteLine (pDocument->pd_ThetaD, Model::sui_Topics);
		csvR.WriteLine (&pDocument->d_R, 1);
		csvTokenTopic.WriteLine (pDocument->pui_TokenTopic, pDocument->ui_DocumentTokens);
		csvTokenTopicSource.WriteLine (pDocument->pc_TokenTopicSource, pDocument->ui_DocumentTokens);
	}
}


// -------------------------------------------------------------------
void Model::LoadLanguageModelState (void)
{
	// save language model ...
	unsigned long ulCount = Model::sui_Topics * Document::sui_Types;
	CsvFile csvPhi ((config)"phi_state_file");
	csvPhi.ReadLine (pd_PhiData, &ulCount);

	// save document Thetas, R, token topics & token topic sources ...
	CsvFile csvTheta ((config)"thetad_state_file");
	CsvFile csvR ((config)"r_state_file");
	CsvFile csvTokenTopic ((config)"token_topic_state_file");
	CsvFile csvTokenTopicSource ((config)"token_topic_source_state_file");

	unsigned long ulTopics = Model::sui_Topics;
	unsigned long ulR = 1;
	ITERATE (Document_dq_t, Document::sdq_Documents, ite)
	{
		Document* pDocument = *ite;
		unsigned long ulDocumentTokens = pDocument->ui_DocumentTokens;

		csvTheta.ReadLine (pDocument->pd_ThetaD, &ulTopics);
		csvR.ReadLine (&pDocument->d_R, &ulR);
		csvTokenTopic.ReadLine (pDocument->pui_TokenTopic, &ulDocumentTokens);
		csvTokenTopicSource.ReadLine (pDocument->pc_TokenTopicSource, &ulDocumentTokens);
	}
}


// -------------------------------------------------------------------
void Model::SaveClusterParams (void)
{
	CsvFile csvTheta ((config)"theta_file", ios_base::app|ios_base::out);
	csvTheta.WriteLine (pd_Theta, Model::sui_Topics);

	CsvFile csvZ ((config)"z_file", ios_base::app|ios_base::out);
	for (unsigned int i = 0; i < 1000; ++ i)
		csvZ.WriteLine (&(pui_TopicLogData [i * Phrase::sui_PhraseCount]),
						Phrase::sui_PhraseCount);
}


// -------------------------------------------------------------------
void Model::SaveLanguageModelParams (int _iIteration)
{
	return;
	// LM averages ...
	/*
	{
		String sPhiFile;
		sPhiFile << (config)"phi_file" << ".avg." << _iIteration / 1000 << "k.csv";
		CsvFile csvPhi (sPhiFile, ios_base::out);

		// compute Phi averages ...
		double dPhiAvg [Document::sui_Types * Model::sui_Topics];
		memset (dPhiAvg, 0, sizeof (double) * Document::sui_Types * Model::sui_Topics);
		for (int i = 0; i < 1000; ++ i)
		{
			for (unsigned long j = 0; j < Document::sui_Types * Model::sui_Topics; ++ j)
				dPhiAvg [j] += ppd_PhiLog [i][j];
		}
		for (unsigned long j = 0; j < Document::sui_Types * Model::sui_Topics; ++ j)
			dPhiAvg [j] /= 1000;
		
		// write out Phi by topic ...
		double dPhi [Document::sui_Types];
		for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
		{
			for (unsigned int w = 0; w < Document::sui_Types; ++ w)
				dPhi [w] = dPhiAvg [t + w * Model::sui_Topics];
			csvPhi.WriteLine (dPhi, Document::sui_Types);
		}
	}
	*/

	// LM snapshot ...
	/*
	{
		String sPhiFile;
		sPhiFile << (config)"phi_file" << ".cut." << _iIteration / 1000 << "k.csv";
		CsvFile csvPhi (sPhiFile, ios_base::out);

		double dPhi [Document::sui_Types];
		for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
		{
			for (unsigned int w = 0; w < Document::sui_Types; ++ w)
				dPhi [w] = pd_PhiData [t + w * Model::sui_Topics];
			csvPhi.WriteLine (dPhi, Document::sui_Types);
		}
	}
	*/
}


// -------------------------------------------------------------------
void Model::Sample (void)
{
	// log run configuration ...
	File fileRunConfig ((config)"run_config_file", ios_base::out);
	time_t now = time (NULL);
	fileRunConfig << ctime (&now) << endl;
	fileRunConfig << endl;
	fileRunConfig << "code version  : 9.0" << endl;
	fileRunConfig << "domain        : " << (config)"domain" << endl;
	fileRunConfig << "clusters      : " << (config)"topic_count" << endl;
	fileRunConfig << endl;
	fileRunConfig << "topic smoothing : " << (config)"topic_smoothing" << endl;
	fileRunConfig << "cluster prior   : " << (config)"cluster_prior" << endl;
	fileRunConfig << "topic prior     : " << (config)"topic_prior" << endl;
	fileRunConfig << "LM prior        : " << (config)"lm_prior" << endl;
	fileRunConfig << endl;
	fileRunConfig << "data path     : " << Path::GetCanonicalPath ((config)"data_path") << endl;
	fileRunConfig << "results path  : " << Path::GetCanonicalPath ((config)"results_path") << endl;
	fileRunConfig << endl;
	fileRunConfig << "phrase cluster iterations    : " << (config)"cluster_iterations" << endl;
	fileRunConfig << "LM initialization iterations : " << (config)"lm_iterations" << endl;
	fileRunConfig << "joint sampling iterations    : " << (config)"joint_iterations" << endl;
	fileRunConfig << "sub iterations               : " << (config)"joint_sub_iterations" << endl;
	fileRunConfig.close ();


	// Check if we're supposed to continue from previous run ...
	if ((config)"start" == "continue")
	{
		LoadClusterState ();
		LoadLanguageModelState ();
	}
	
	// ----------------------------------------------------
	clock_t start = clock ();
	int iClusterLogIte = (config)"log_cluster_iterations_at";
	int iClusterIterations = (config)"cluster_iterations";
	int iSaveClusterStateAt = (config)"save_cluster_state_at";

	for (int i = 1; i <= iClusterIterations; ++ i)
	{
		if (0 == (i % iClusterLogIte))
			cout << "  cluster : " << i << endl;
		if (0 == (i % iSaveClusterStateAt))
			SaveClusterState ();
		if (0 == (i % 1000))
			SaveClusterParams ();

		SampleClusters ();

		// maintain phrase topic log ...
		int p = 0;
		unsigned int uiOffset = (i % 1000) * Phrase::sui_PhraseCount;
		ITERATE (Phrase_dq_t, Phrase::sdq_Phrases, ite)
			pui_TopicLogData [uiOffset + (p++)] = (*ite)->ui_Topic;
	}

	cout << "  cluster : " << iClusterIterations << "  " 
		 << (clock () - start) / (double)CLOCKS_PER_SEC << endl;
	cout << endl;


	if ((config)"use_gold_clusters" == "true")
	{
		cout << "[WARNING] using gold cluster information." << endl;
		String_dq_t	dqClusterIndices;
		File::ReadLines ((config)"gold_cluster_index_file", dqClusterIndices);
		for (unsigned int i = 0; i < Phrase::sui_PhraseCount; ++ i)
		{
			Phrase::sdq_Phrases [i]->ui_Topic = (int) dqClusterIndices [i];
			cout << Phrase::sdq_Phrases [i]->ui_Topic << endl;
		}
	}

	// ----------------------------------------------------
	start = clock ();
	int iLMLogIte = (config)"log_lm_iterations_at";
	int iLMIterations = (config)"lm_iterations";
	int iSaveLMStateAt = (config)"save_lm_state_at";

	for (int i = 1; i <= iLMIterations; ++ i)
	{
		if (0 == (i % iLMLogIte))
			cout << "  lm : " << i << endl;
		if (0 == (i % iSaveLMStateAt))
			SaveLanguageModelState ();
		if (0 == (i % 1000))
			SaveLanguageModelParams (iClusterIterations + i);

		SampleLanguageModels ();
	}

	cout << "  lm : " << iLMIterations << "  " 
		 << (clock () - start) / (double)CLOCKS_PER_SEC << endl;
	cout << endl;

	
	// ----------------------------------------------------
	start = clock ();
	int iJointLogIte = (config)"log_joint_iterations_at";
	int iJointIterations = (config)"joint_iterations";
	int iJointSubIterations = (config)"joint_sub_iterations";
	int iSaveJointStateAt = (config)"save_joint_state_at";

	int i = 0;
	while (i <= iJointIterations)
	{
		if (0 != i)
		{
			if (0 == (i % iJointLogIte))
				cout << "  joint : " << i << endl;
			if (0 == (i % iSaveJointStateAt))
			{
				SaveClusterState ();
				SaveLanguageModelState ();
			}
			if (0 == (i % 1000))
			{
				SaveClusterParams ();
				SaveLanguageModelParams (iClusterIterations + iLMIterations + i);

				d_TopicSmoothing *= d_TopicSmoothingScaling;
				d_Alpha *= d_AlphaScaling;
				d_AlphaD *= d_AlphaDScaling;
				d_Beta *= d_BetaScaling;
			}
		}

		for (int j = 0; j < iJointSubIterations; ++ j)
		{
			SampleClusters (true);
	
			// maintain phrase topic log ...
			int p = 0;
			unsigned int uiOffset = ((i+j) % 1000) * Phrase::sui_PhraseCount;
			ITERATE (Phrase_dq_t, Phrase::sdq_Phrases, ite)
				pui_TopicLogData [uiOffset + (p++)] = (*ite)->ui_Topic;
		}
		for (int j = 0; j < iJointSubIterations; ++ j)
		{
			if (i >= (iJointSubIterations - (int)sui_TopicSamples))
				sb_PredictTestDocumentProperties = true;
			SampleLanguageModels ();
		}
		
		i += iJointSubIterations;
	}

	cout << "  joint : " << iJointIterations << "  " 
		 << (clock () - start) / (double)CLOCKS_PER_SEC << endl;
	cout << endl;
	cout << "done." << endl;
	cout << endl;
}



// -------------------------------------------------------------------
void Model::SampleClusters (bool _bLinkToDocuments)
{
	// keep a count of topic occurrences so we can sample Theta later 
	double dUpdatedThetaParam [Model::sui_Topics];
	for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
		dUpdatedThetaParam [i] = d_Alpha;

	gsl_ran_shuffle (p_GslRng, pp_PhrasesForSampling, Phrase::sui_PhraseCount, sizeof (Phrase*));

	// sample topics for each phrase ... 
	for (unsigned int n = 0; n < Phrase::sui_PhraseCount; ++ n)
	{
		Phrase* pPhrase = pp_PhrasesForSampling [n];
		double* pvdSimilarity = pPhrase->pd_SimilarityVector;

		double dProbIn [Phrase::sui_PhraseCount];
		double dProbOut [Phrase::sui_PhraseCount];
		for (unsigned int i = 0; i < Phrase::sui_PhraseCount; ++ i)
		{
			dProbIn [i] = 2 * (pvdSimilarity [i]);
			dProbOut [i] = 2 - (dProbIn [i]);
		}

		// compute the multinomial probabilities for the topic of this phrase 
		double dPDF [Model::sui_Topics];
		for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
			dPDF [i] = pd_Theta [i];

		unsigned int j = 0;
		ITERATE (Phrase_dq_t, Phrase::sdq_Phrases, iteOther)
		{
			Phrase* pOther = *iteOther;
			if (pOther == pPhrase)
			{
				++ j;	// [SB]{BUG FIX! 2008/07/06} 
				continue;
			}
			for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
				dPDF [i] *= (i == pOther->ui_Topic)? dProbIn [j] : dProbOut [j];
			++ j;
		}

		// compute the phrase topic probabilities based on the document token topics 
		unsigned int uiTopicAnnot [Model::sui_Topics];
		double dDocumentTopicProb [Model::sui_Topics];
		for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
			dDocumentTopicProb [i] = 1;

		if (true == _bLinkToDocuments)
		{
			ITERATE (PhraseDocument_dq_t, pPhrase->dq_PhraseDocuments, iteDoc)
			{
				Document* pDoc = (Document*) *iteDoc;
				if (true == pDoc->b_TestDocument)
					continue;

				// find the topic annotations for the document minus the current phrase ...
				unsigned int iTopicAnnotCount = 0;
				memset (uiTopicAnnot, 0, sizeof (unsigned int) * Model::sui_Topics);
				for (unsigned int p = 0; p < pDoc->ui_DocumentPhrases; ++ p)
				{
					Phrase* pOther = pDoc->pp_PhraseAnnotation [p];
					if (pOther == pPhrase)
						continue;
					if (0 == uiTopicAnnot [pOther->ui_Topic])
						++ iTopicAnnotCount;
					uiTopicAnnot [pOther->ui_Topic] = 1;
				}

				// for each potential topic for current phrase, 
				for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
				{
					unsigned int uiPrevious = uiTopicAnnot [t];
					uiTopicAnnot [t] = 1;
					double dNormalization = (1 == uiPrevious)? 
												iTopicAnnotCount + d_TopicSmoothing * Model::sui_Topics: 
												iTopicAnnotCount + 1 + d_TopicSmoothing * Model::sui_Topics;

					for (unsigned int x = 0; x < Model::sui_Topics; ++ x)
					{
						if (0 == pDoc->pul_PhraseTopicTokenCounts [x])
							continue;
						dDocumentTopicProb [t] *= pow (((d_TopicSmoothing + uiTopicAnnot [x]) / dNormalization), 
														(double)pDoc->pul_PhraseTopicTokenCounts [x]);
					}

					uiTopicAnnot [t] = uiPrevious;
				}
			}

			// normalize dDocumentTopicProb 
			double dSum = 0;
			for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
				dSum += dDocumentTopicProb [i];
			if (0 != dSum)
			{
				for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
					dDocumentTopicProb [i] /= dSum;
			}
			else
			{
				for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
					dDocumentTopicProb [i] = 1;
			}
		}


		// sample the phrase's topic from the multinomial computed above 
		double dCDF [Model::sui_Topics];
		dCDF [0] = dPDF [0] * dDocumentTopicProb [0];
		for (unsigned int i = 1; i < Model::sui_Topics; ++ i)
			dCDF [i] = dCDF [i-1] + dPDF [i] * dDocumentTopicProb [i];
		for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
			pPhrase->pd_TopicProbability [i] = dPDF [i] / dCDF [Model::sui_Topics - 1];

		double dUniformRnd = dCDF [Model::sui_Topics - 1] * gsl_rng_uniform (p_GslRng);
		for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
		{
			if (dUniformRnd > dCDF [i])
				continue;
			pPhrase->ui_Topic = i;
			++ dUpdatedThetaParam [i];
			break;
		}
	}

	// sample Theta ...
	gsl_ran_dirichlet (p_GslRng, Model::sui_Topics, dUpdatedThetaParam, pd_Theta);
}



// -------------------------------------------------------------------
void Model::SampleLanguageModels (void)
{
	double dTypeTopicCountData [Model::sui_Topics * Document::sui_Types];
	for (unsigned long i = 0; i < Model::sui_Topics * Document::sui_Types; ++ i)
		dTypeTopicCountData [i] = d_Beta;
	double* ppdTypeTopicCounts [Model::sui_Topics];
	for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
		ppdTypeTopicCounts [i] = &(dTypeTopicCountData [i * Document::sui_Types]);


	// unlock starting mutexes 
	for (int i = 0; i < i_ThreadCount; ++ i)
		pthread_mutex_unlock (&p_ThreadInfo [i].mtx_Start);
	// lock ending mutexes 
	for (int i = 0; i < i_ThreadCount; ++ i)
		pthread_mutex_lock (&p_ThreadInfo [i].mtx_Done);

	// compute overall type topic counts 
	for (int i = 0; i < i_ThreadCount; ++ i)
	{
		ThreadInfo* pInfo = &(p_ThreadInfo [i]);
		for (unsigned long j = 0; j < Model::sui_Topics * Document::sui_Types; ++ j)
			dTypeTopicCountData [j] += pInfo->pd_TypeTopicCountData [j];
	}

	// sample language models for each topic ...
	double dPhiSample [Document::sui_Types];
	for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
	{
		gsl_ran_dirichlet (p_GslRng, Document::sui_Types, ppdTypeTopicCounts [t], dPhiSample);
		for (unsigned int w = 0; w < Document::sui_Types; ++ w)
			ppd_Phi [w][t] = dPhiSample [w];
	}
}


// -------------------------------------------------------------------
void* Model::SampleDocumentLanguageModels (void* _pParam)
{
	ThreadInfo* pThreadInfo = (ThreadInfo*) _pParam;
	
	double* ppdTypeTopicCounts [Model::sui_Topics];
	for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
		ppdTypeTopicCounts [i] = &(pThreadInfo->pd_TypeTopicCountData [i * Document::sui_Types]);

	while (false == pThreadInfo->b_Done)
	{
		// wait till SampleLanguageModels unlocks start ...
		pthread_mutex_lock (&pThreadInfo->mtx_Start);

		// initialize 
		memset (pThreadInfo->pd_TypeTopicCountData, 0, 
				sizeof (double) * Model::sui_Topics * Document::sui_Types);

		// randomly shuffle the sampling sequence ...
		gsl_ran_shuffle (pThreadInfo->p_GslRng, 
						 pThreadInfo->pp_DocumentsForSampling, 
						 pThreadInfo->l_DocumentsForSampling,
						 sizeof (Document*));

		// sample 
		for (long n = 0; n < pThreadInfo->l_DocumentsForSampling; ++ n)
		{
			Document* pDoc = pThreadInfo->pp_DocumentsForSampling [n];
			if ((true == pDoc->b_TestDocument) && (false == sb_PredictTestDocumentProperties))
				continue;

			// compute document topic annotation from phrase annotation 
			// memset (pDoc->pd_TopicAnnotation, 0, sizeof (double) * Model::sui_Topics);
			for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
				pDoc->pd_TopicAnnotation [t] = d_TopicSmoothing;
			for (unsigned int i = 0; i < pDoc->ui_DocumentPhrases; ++ i)
			{
				Phrase* pPhrase = pDoc->pp_PhraseAnnotation [i];
				pDoc->pd_TopicAnnotation [pPhrase->ui_Topic] = 1;
			}
			double dTopicAnnotationCount = 0;
			for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
				dTopicAnnotationCount += pDoc->pd_TopicAnnotation [t];


			// randomly shuffle token sampling sequence ...
			gsl_ran_shuffle (pThreadInfo->p_GslRng,
							 pDoc->pui_TokenIndex,
							 pDoc->ui_DocumentTokens,
							 sizeof (unsigned int));


			// sample token topics & type topic counts ...
			for (unsigned int wi = 0; wi < pDoc->ui_DocumentTokens; ++ wi)
			{
				unsigned int w = pDoc->pui_TokenIndex [wi];

				unsigned int iType = pDoc->pui_TokenType [w];
				double* pTopicProbability = (1 == pDoc->pc_TokenTopicSource [w])? 
												pDoc->pd_TopicAnnotation :
												pDoc->pd_ThetaD;

				double dCDF [Model::sui_Topics];

				// 50s
				double* pdPhi = pThreadInfo->ppd_Phi [iType];
				dCDF [0] = pdPhi [0] * pTopicProbability [0];
				for (unsigned int t = 1; t < Model::sui_Topics; ++ t)
					dCDF [t] = dCDF [t-1] + pdPhi [t] * pTopicProbability [t];
				if (0 == dCDF [Model::sui_Topics - 1])
				{
					for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
						dCDF [t] = t + 1;
				}

				// 26s
				double dUniformRnd = dCDF [Model::sui_Topics - 1] * 
									 gsl_rng_uniform (pThreadInfo->p_GslRng);

				for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
				{
					if (dCDF [t] < dUniformRnd)
						continue;
					pDoc->pui_TokenTopic [w] = t;
					if (false == pDoc->b_TestDocument)
						++ ppdTypeTopicCounts [t][iType];
					break;
				}
			}


			// compute count of tokens for each topic where token topic source is ThetaD
			double dTokenCountByTopic [Model::sui_Topics];
			for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
				dTokenCountByTopic [t] = pThreadInfo->d_AlphaD;

			for (unsigned int w = 0; w < pDoc->ui_DocumentTokens; ++ w)
				dTokenCountByTopic [pDoc->pui_TokenTopic [w]] += (0 == pDoc->pc_TokenTopicSource [w]);

			// sample ThetaD	
			gsl_ran_dirichlet (pThreadInfo->p_GslRng, 
							   Model::sui_Topics, 
							   dTokenCountByTopic,
							   pDoc->pd_ThetaD);
			if ((true == pDoc->b_TestDocument) && (true == sb_PredictTestDocumentProperties))
			{
				for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
					pDoc->pd_SumThetaD [t] += pDoc->pd_ThetaD [t];
			}


			// sample token topic source	
			memset (pDoc->pul_PhraseTopicTokenCounts, 0, sizeof (unsigned long) * Model::sui_Topics);

			unsigned int uiTopicsFromPhraseAnnotation = 0;
			if (0 == dTopicAnnotationCount)
			{
				for (unsigned int w = 0; w < pDoc->ui_DocumentTokens; ++ w)
					pDoc->pc_TokenTopicSource [w] = 0;
			}
			else
			{
				for (unsigned int wi = 0; wi < pDoc->ui_DocumentTokens; ++ wi)
				{
					unsigned int w = pDoc->pui_TokenIndex [wi];
					unsigned int uiTopic = pDoc->pui_TokenTopic [w];
					double dP0 = (1 - pDoc->d_R) * pDoc->pd_ThetaD [uiTopic];
					double dP1 = pDoc->d_R * pDoc->pd_TopicAnnotation [uiTopic] 
									/ dTopicAnnotationCount;

					if (gsl_rng_uniform (pThreadInfo->p_GslRng) > dP0 / (dP0 + dP1))
					{
						pDoc->pc_TokenTopicSource [w] = 1;
						++ uiTopicsFromPhraseAnnotation;
						++ pDoc->pul_PhraseTopicTokenCounts [pDoc->pui_TokenTopic [w]];
					}
					else
						pDoc->pc_TokenTopicSource [w] = 0;
				}
			}

			// sample R	
			pDoc->d_R = gsl_ran_beta (pThreadInfo->p_GslRng, 
									  1 + uiTopicsFromPhraseAnnotation, 
									  1 + pDoc->ui_DocumentTokens - uiTopicsFromPhraseAnnotation);
		}

		// unlock 
		pthread_mutex_unlock (&pThreadInfo->mtx_Done);
	}

	delete pThreadInfo;
	return NULL;
}



