#include "nlp_filesystem.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
using namespace std;


// --------------------------------------------------------------------
bool Path::Exists (const char* _zPath)
{
	return Access (_zPath, F_OK);
}

bool Path::Readable (const char* _zPath)
{
	return Access (_zPath, R_OK);
}

bool Path::Writable (const char* _zPath)
{
	return Access (_zPath, W_OK);
}

bool Path::Runable (const char* _zPath)
{
	return Access (_zPath, X_OK);
}

// --------------------------------------------------------------------
bool Path::Access (const char* _zPath, int _iMode)
{
	if (0 == access (_zPath, _iMode))
		return true;

	if (EACCES == errno)
	{
		cerr << "[WARNING]  Access denied to path '" << _zPath 
			 << "'." << endl;
		return false;
	}
	if (ELOOP == errno)
	{
		cerr << "[ERROR]  Too many symbolic links were encountered in resolving path '"
			 << _zPath << "'." << endl;
		return false;
	}
	if (ENAMETOOLONG == errno)
	{
		cerr << "[ERROR]  Path name too long : '" << _zPath << "'." << endl;
		return false;
	}
	if (ENOTDIR == errno)
	{
		cerr << "[ERROR]  A component used as a directory in the path is not "
				"in fact a directory. Path : '" << _zPath << "'." << endl;
		return false;
	}
	if ((ENOENT == errno) && (F_OK != _iMode))
	{
		cerr << "[ERROR]  File does not exist, or is a dangling link. Path : '"
			 << _zPath << "'." << endl;
		return false;
	}
	if (EFAULT == errno)
	{
		cerr << "[ERROR]  Path points outside process' accessible address space."
			 << endl;
		return false;
	}
	if (EINVAL == errno)
	{
		cerr << "[ERROR]  Invalid mode parameter specified to Path::Access. Mode = "
			 << _iMode << "." << endl;
		return false;
	}
	if (EIO == errno)
	{
		cerr << "[ERROR]  An I/O error occurred while checking access to path '"
		 	 << _zPath << "'." << endl;
		return false;
	}
	if (ENOMEM == errno)
	{
		cerr << "[ERROR]  Insufficient kernel memory while checking access to path '"
			 << _zPath << "'." << endl;
		return false;
	}

	return false;
}


// --------------------------------------------------------------------
bool Path::CreatePath (const char* _zPath, mode_t _iMode)
{
	// if (0 == access (_zPath, F_OK))
		// return true;
	if (0 == mkdir (_zPath, _iMode))
		return true;

	int iErrorNo = errno;
	cerr << "[ERROR]  Failed to create path " << _zPath 
		 << " . Error : " << strerror (iErrorNo) << endl;
	return false;
}


// --------------------------------------------------------------------
bool Path::RemovePath (const char* _zPath)
{
	// if path doesn't exist, we don't need to do anything ...
	if (0 != access (_zPath, F_OK))
		if (ENOENT == errno)
			return true;

	// delete path ...
	if (0 == rmdir (_zPath))
		return true;

	int iErrorNo = errno;
	cerr << "[ERROR]  Failed to delete path " << _zPath 
		 << " . Error : " << strerror (iErrorNo) << endl;
	return false;
}


// --------------------------------------------------------------------
bool Path::MovePath (const char* _zPath)
{
	return false;
}


// --------------------------------------------------------------------
bool Path::MoveFile (const char* _zFileName)
{
	return false;
}


// --------------------------------------------------------------------
bool Path::RemoveFile (const char* _zFileName)
{
	if (0 == unlink (_zFileName))
		return true;

	int iErrorNo = errno;
	cerr << "[ERROR]  Failed to delete file " << _zFileName 
		 << " . Error : " << strerror (iErrorNo) << endl;
	return false;
}


// --------------------------------------------------------------------
bool Path::GetFileList (const char* _zPath, String_dq_t& _rdqFileNames)
{
	DIR* pDirectory = opendir (_zPath);
	if (NULL == pDirectory)
	{
		if (EACCES == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', permission denied." << endl;
		else if (EMFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many file descriptors in use by process." << endl;
		else if (ENFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many files are currently open in the system." << endl;
		else if (ENOENT == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', directory does not exist." << endl;
		else if (ENOMEM == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', insufficient memory." << endl;
		else if (ENOTDIR == errno) 
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', this is not a directory." << endl;
		else
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', unknown error." << endl;
		return false;
	}

	dirent* pFile = readdir (pDirectory);
	if (NULL == pFile)
	{
		if (EACCES == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath 
				 << "', permission denied." << endl;
		else if (EMFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many file descriptors in use by process." << endl;
		else if (ENFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many files are currently open in the system." << endl;
		else if (ENOENT == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', directory does not exist." << endl;
		else if (ENOMEM == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', insufficient memory." << endl;
		else if (ENOTDIR == errno) 
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', this is not a directory." << endl;
		else
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', unknown error." << endl;
	}

	while (NULL != pFile)
	{
		if ((0 != strcmp (".", pFile->d_name)) &&
			(0 != strcmp ("..", pFile->d_name)))
			_rdqFileNames.push_back (pFile->d_name);

		pFile = readdir (pDirectory);
	}
	closedir (pDirectory);

	return true;
}


// --------------------------------------------------------------------
bool Path::GetPathList (const char* _zPath, String_dq_t& _rdqFileNames)
{
	DIR* pDirectory = opendir (_zPath);
	if (NULL == pDirectory)
	{
		if (EACCES == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', permission denied." << endl;
		else if (EMFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many file descriptors in use by process." << endl;
		else if (ENFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many files are currently open in the system." << endl;
		else if (ENOENT == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', directory does not exist." << endl;
		else if (ENOMEM == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', insufficient memory." << endl;
		else if (ENOTDIR == errno) 
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', this is not a directory." << endl;
		else
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', unknown error." << endl;
		return false;
	}

	dirent* pFile = readdir (pDirectory);
	if (NULL == pFile)
	{
		if (EACCES == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath 
				 << "', permission denied." << endl;
		else if (EMFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many file descriptors in use by process." << endl;
		else if (ENFILE == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', too many files are currently open in the system." << endl;
		else if (ENOENT == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', directory does not exist." << endl;
		else if (ENOMEM == errno)
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', insufficient memory." << endl;
		else if (ENOTDIR == errno) 
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', this is not a directory." << endl;
		else
			cerr << "[ERROR]  Failed to open directory '" << _zPath
				 << "', unknown error." << endl;
	}

	while (NULL != pFile)
	{
		if ((0 != strcmp (".", pFile->d_name)) &&
			(0 != strcmp ("..", pFile->d_name)))
		{
			String sPath;
			if ('/' == _zPath [strlen (_zPath) - 1])
				sPath << _zPath << pFile->d_name;
			else
				sPath << _zPath << '/' << pFile->d_name;
			_rdqFileNames.push_back (sPath);
		}

		pFile = readdir (pDirectory);
	}
	closedir (pDirectory);

	return true;
}


// --------------------------------------------------------------------
String Path::GetCanonicalPath (const char* _zPath)
{
	long lBufferLength = 0;
	long lPathLength = 0;
	char* pCanonicalPath = NULL;
	do
	{
		lBufferLength += 1000;
		pCanonicalPath = (char*) realloc (pCanonicalPath, lBufferLength);
		lPathLength = readlink (_zPath, pCanonicalPath, lBufferLength);
	}
	while ((lPathLength == lBufferLength) && (-1 != lPathLength));
	
	if (-1 == lPathLength)
	{
		int iError = errno;
		free (pCanonicalPath);

		// if the path doesn't have any symbolic links in it, readlink returns EINVAL! 
		if (EINVAL != iError)
			cerr << "[ERROR]  Canonicalization failed for path " << _zPath 
				 << " , Error : " << strerror (iError) << endl;
		return _zPath;
	}

	pCanonicalPath [lPathLength] = '\0';
	String sCanonicalPath (pCanonicalPath);
	free (pCanonicalPath);

	return sCanonicalPath;
}
